/*
 * Copyright 2000 by Docuverse
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Docuverse ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you
 * entered into with Docuverse.
 */

package com.docuverse.min;

//import java.io.File;
import java.io.InputStream;
import java.io.Reader;
import java.io.IOException;
import java.io.EOFException;
//import java.io.CharConversionException;
import java.io.UnsupportedEncodingException;
//import java.net.URL;
//import java.net.MalformedURLException;
//import java.util.Locale;
import org.xml.sax.InputSource;
import org.xml.sax.EntityResolver;
import org.xml.sax.DTDHandler;
import org.xml.sax.DocumentHandler;
import org.xml.sax.ErrorHandler;
import org.xml.sax.HandlerBase;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/** 
Minimal XML parser implementation.

@author  Don Park
@version 1.0
*/

public class Parser implements org.xml.sax.Locator, org.xml.sax.Parser
{
// compile time flags

	private final static boolean JAVA12 = false;

// constants

	final private static int		BYTE_READ_LENGTH = 8 * 1024;
	final private static int		BYTE_BUFFER_SIZE = 8 * 1024 + 5;
	final private static int		CHAR_READ_LENGTH = 4 * 1024;
	final private static int		CHAR_BUFFER_SIZE = 4 * 1024 + 1;
	final private static int		TEMP_BUFFER_SIZE = 4 * 1024;
	final private static char		END_OF_DATA = '\0';
	final private static int		TEMP_NEWLINE = 0;
	final private static int		TEMP_CHAR1 = 1;
	final private static int		TEMP_CHAR2 = 2;
	final private static int		TEMP_START = 3;

// instance variables

	private byte[]					byteBuf;	// shared buffer for reading in bytes
	private char[]					saveBuf;	// buffer saving stuff (name, whitespace)
    private char[]					readBuf;
	private org.xml.sax.HandlerBase	defHandler;
	private EmptyAttributeList		attributes;

    //private Locale					msgLocale;
	private EntityResolver			pidHandler;
	private DocumentHandler			docHandler;
	private ErrorHandler			errHandler;

	private String					publicId;
	private String					systemId;
	private Reader					reader;
	private boolean					readerClosable;
	private int						savePos;
    private int						readPos;
    private int						line;
    private NameTable				nameTable;
    
		/** Default constructor.
		 */
        public
    Parser()
    {
		byteBuf = new byte[BYTE_BUFFER_SIZE];
		readBuf = new char[CHAR_BUFFER_SIZE]; // static size
		saveBuf = new char[TEMP_BUFFER_SIZE]; // dynamic size
		saveBuf[TEMP_NEWLINE] = '\n';

		defHandler = new HandlerBase();
		attributes = new EmptyAttributeList();
		docHandler = defHandler;
		pidHandler = defHandler;
		errHandler = defHandler;
	}

// org.xml.sax.Parser methods


	  /**
		* Allow an application to register a custom entity resolver.
		*
		* <p>If the application does not register an entity resolver, the
		* SAX parser will resolve system identifiers and open connections
		* to entities itself (this is the default behaviour implemented in
		* HandlerBase).</p>
		*
		* <p>Applications may register a new or different entity resolver
		* in the middle of a parse, and the SAX parser must begin using
		* the new resolver immediately.</p>
		*
		* @param resolver The object for resolving entities.
		* @see EntityResolver
		* @see HandlerBase
		*/
		public void
	setEntityResolver(EntityResolver resolver)
	{
		pidHandler = resolver != null ? resolver : defHandler;
	}

	  /**
		* Allow an application to register a DTD event handler.
		*
		* <p>If the application does not register a DTD handler, all DTD
		* events reported by the SAX parser will be silently
		* ignored (this is the default behaviour implemented by
		* HandlerBase).</p>
		*
		* <p>Applications may register a new or different
		* handler in the middle of a parse, and the SAX parser must
		* begin using the new handler immediately.</p>
		*
		* @param handler The DTD handler.
		* @see DTDHandler
		* @see HandlerBase
		*/
		public void
	setDTDHandler(DTDHandler handler)
	{
		// ignored
	}

	  /**
		* Allow an application to register a document event handler.
		*
		* <p>If the application does not register a document handler, all
		* document events reported by the SAX parser will be silently
		* ignored (this is the default behaviour implemented by
		* HandlerBase).</p>
		*
		* <p>Applications may register a new or different handler in the
		* middle of a parse, and the SAX parser must begin using the new
		* handler immediately.</p>
		*
		* @param handler The document handler.
		* @see DocumentHandler
		* @see HandlerBase
		*/
		public void
	setDocumentHandler(DocumentHandler handler)
	{
		docHandler = handler != null ? handler : defHandler;
	}

	  /**
		* Allow an application to register an error event handler.
		*
		* <p>If the application does not register an error event handler,
		* all error events reported by the SAX parser will be silently
		* ignored, except for fatalError, which will throw a SAXException
		* (this is the default behaviour implemented by HandlerBase).</p>
		*
		* <p>Applications may register a new or different handler in the
		* middle of a parse, and the SAX parser must begin using the new
		* handler immediately.</p>
		*
		* @param handler The error handler.
		* @see ErrorHandler
		* @see SAXException
		* @see HandlerBase
		*/
		public void
	setErrorHandler(ErrorHandler handler)
	{
		errHandler = handler != null ? handler : defHandler;
	}

	  /**
		* Parse a Minimal XML document.
		*
		* <p>A standard SAX method.</p>
		* <p>The application can use this method to instruct the SAX parser
		* to begin parsing a Minimal XML document from any valid input
		* source (a character stream, a byte stream, or a URI).</p>
		*
		* <p>Applications may not invoke this method while a parse is in
		* progress (they should create a new Parser instead for each
		* additional Minimal XML document).  Once a parse is complete, an
		* application may reuse the same Parser object, possibly with a
		* different input source.</p>
		*
		* @param source The input source for the top-level of the
		*        Minimal XML document.
		* @exception org.xml.sax.SAXException Any SAX exception, possibly
		*            wrapping another exception.
		* @exception java.io.IOException An IO exception from the parser,
		*            possibly from a byte stream or character stream
		*            supplied by the application.
		* @see org.xml.sax.InputSource
		* @see #parse(java.lang.String)
		* @see #setEntityResolver
		* @see #setDTDHandler
		* @see #setDocumentHandler
		* @see #setErrorHandler
		*/
		public void
	parse(InputSource source)
		throws SAXException, IOException
	{
		parse(source, false);
	}

	  /**
		* Parse a Minimal XML document or fragment.
		*
		* <p>The application can use this method to instruct the SAX parser
		* to begin parsing a Minimal XML document or fragment from any valid
		* input source (a character stream, a byte stream, or a URI).</p>
		*
		* <p>Applications may not invoke this method while a parse is in
		* progress (they should create a new Parser instead for each
		* additional Minimal XML document or fragment).  Once a parse is
		* complete, an application may reuse the same Parser object, possibly
		* with a different input source.</p>
		*
		* @param source The input source for the Minimal XML document or
		*				fragment.
		* @param fragment specifies whether the content is a single document
		*	with a single element at root or a fragment which may have
		*	multiple element at root.
		* @exception org.xml.sax.SAXException Any SAX exception, possibly
		*            wrapping another exception.
		* @exception java.io.IOException An IO exception from the parser,
		*            possibly from a byte stream or character stream
		*            supplied by the application.
		* @see org.xml.sax.InputSource
		* @see #parse(java.lang.String)
		* @see #setEntityResolver
		* @see #setDTDHandler
		* @see #setDocumentHandler
		* @see #setErrorHandler
		*/
		public void
	parse(InputSource source, boolean fragment)
		throws SAXException, IOException
	{
		Reader		r;
		InputStream	in;

		systemId = source.getSystemId();
		publicId = source.getPublicId();

		if ((r = source.getCharacterStream()) != null)
			parse(r, fragment, false);
		else if ((in = source.getByteStream()) != null)
			parse(createReader(in, false), fragment, true);
		else
		    throw new IOException("can't open stream.");
		//parse((systemId != null ? systemId : publicId), fragment);
	}


// org.xml.sax.Locator methods

		public String
	getPublicId()
	{
		return publicId;
	}

		public String
	getSystemId()
	{
		return systemId;
	}

		public int
	getLineNumber()
	{
		return line;
	}

		public int
	getColumnNumber()
	{
		return -1; // not available
	}

// overridable methods

		protected Reader
	createReader(InputStream in, boolean closable)
		throws IOException
    {
		int n = in.read(byteBuf, 0, BYTE_READ_LENGTH);
		if (n < 2)
			throw new IOException("input too short");

		int b0 = byteBuf[0] & 0x0FF;
		int b1 = byteBuf[1] & 0x0FF;

			// big-endian UTF-16 without byte-order mark?
		if (b0 == 0)
			return new UTF16B(in, closable, byteBuf, 0, n);
			// little-endian UTF-16 without byte-order mark?
		else if (b1 == 0)
			return new UTF16L(in, closable, byteBuf, 0, n);
			// big-endian UTF-16 with byte-order mark?
		else if (b0 == 0x0FE && b1 == 0x0FF)
			return new UTF16B(in, closable, byteBuf, 2, n - 2);
			// little-endian UTF-16 with byte-order mark?
		else if (b0 == 0x0FF && b1 == 0x0FE)
			return new UTF16L(in, closable, byteBuf, 2, n - 2);
		else
			return new UTF8(in, closable, byteBuf, 0, n);
	}

		protected void
	closeReader()
    {
		try
		{
			if (reader != null)
			{
				if (readerClosable)
					reader.close();
		
				reader = null;
			}
		}
		catch (IOException e)
		{
		}
    }

    public void parse(String systemId) throws SAXException, IOException {
	 parse(new InputSource(systemId));
    }

		/**
		Primary parse method.  Parse the content of given <code>java.io.Reader</code>
		instance as Minimal XML and emit SAX events to installed SAX handlers.

		@param in character stream containing the content to be parsed.
		@param fragment specifies whether the content is a single document
						with a single element at root or a fragment which
						may have multiple element at root.
		@param closable specifies whether the given reader should be closed
				automatically before returning.
		*/
		public void
	parse(Reader in, boolean fragment, boolean closable)
		throws SAXException, IOException
    {
		reader = in;
		readerClosable = closable;

		savePos = TEMP_START;
		readPos = 0;
		line = 1;
		readBuf[0] = END_OF_DATA;
		nameTable = new NameTable();

		docHandler.setDocumentLocator(this);
		try
		{
			for (;;)
			{
				parseDocument();
				if (!fragment)
					break;
			}
		}
		catch (EOFException e)
		{
		}
		catch (UnsupportedEncodingException e)
		{
			error("Unsupported character encoding type: must be UTF-8 or UTF-16");
		}
		//catch (CharConversionException e)
		//{
		//	error("invalid character encoding");
		//}
		catch (Exception e)
		{
			e.printStackTrace();
		}
		finally
		{
			closeReader();
		}
	}

// private methods

		private final void
	parseDocument()
		throws SAXException, IOException
	{
		// [1] document ::= S* element S*

		docHandler.startDocument();
		ignoreWhitespace("root missing");
		if (readBuf[readPos++] != '<')
			error("illegal character; expecting '<'");

		char c = readBuf[readPos];
		if (c == END_OF_DATA)
		{
				// refill the buffer to the rim
			int readLen = reader.read(readBuf, 0, CHAR_READ_LENGTH);
			if (readLen < 0)
				throw new EOFException("EOF in name");
			readBuf[readLen] = END_OF_DATA;
			c = readBuf[readPos = 0];
		}
		if (c == '/')
			error("document element's end tag before start tag");

		parseElement();
		ignoreWhitespace(null);
		docHandler.endDocument();
	}

        private final void
    parseElement()
		throws SAXException, IOException
    {
		char[]	rbuf = readBuf;
		char[]	sbuf = saveBuf;
		int		rpos = readPos;
		int		spos;
		char	c;
		char	cx;
		int		len;
		Name	name;
		boolean	empty;

		// [2] element ::= STag content ETag
		// [4] ETag ::= '</' Chars '>'
		// [5] content ::= (element | WS)* | (CharData | CharRef)*
		// [6] Chars ::= [^<>&/]+
		// [7] CharData ::= [^<>&]
		// [8] CharRef ::= '&#' [0-9]+ ';'
		// [9] WS ::= (#32 | #9 | #13 | #10)

		// Caller already got '<'
		// Caller made sure next character is not '/'

		// [3] STag ::= '<' Chars '>'
		//                  ^
		//=======================================================
		// Get tag name into 'name'
		// spos is the start of name

		spos = savePos;		// name position
        for (;;)
        {
			c = rbuf[rpos++];

				// end of start tag
				// normal loop exit
			if (c == '>')
				break;

			else if (c == '<' || c == '&' || c == '/')
			{
				readPos = rpos;
				error("illegal character in name");
			}
			else if (c == END_OF_DATA)
			{
				if ((len = reader.read(rbuf, 0, CHAR_READ_LENGTH)) < 0)
					throw new EOFException("EOF in name");
				rbuf[len] = END_OF_DATA;
				rpos = 0;
				c = rbuf[rpos++];
				if (c == '>')
					break;
				else if (c == '<' || c == '&' || c == '/')
				{
					readPos = rpos;
					error("illegal character in name");
				}
			}
            if (spos >= sbuf.length)
            {
				sbuf = new char[sbuf.length * 2];
                System.arraycopy(saveBuf, 0, sbuf, 0, saveBuf.length);
				saveBuf = sbuf;
            }
            sbuf[spos++] = c;
        }
		name = nameTable.getName(saveBuf, savePos, spos - savePos);
		if (name.start == savePos)
			savePos = spos;

		// got the name in 'name'
		//-------------------------------------------------------

		docHandler.startElement(name.value, attributes);

		//=======================================================
		// Parse leading whitespaces in content if any
		// We don't know whether to report the whitespaces
		// or not until end of the whitespaces is found

		spos = savePos;		// whitespace position
		for (;;)
		{
				// expand save buffer on demand
            if (spos >= sbuf.length)
            {
				sbuf = new char[sbuf.length * 2];
                System.arraycopy(saveBuf, 0, sbuf, 0, saveBuf.length);
				saveBuf = sbuf;
            }

				// get the next character
			c = rbuf[rpos++];
			if (c == ' ' || c == '\t')
			{
				sbuf[spos++] = c;
			}
			else if (c == '\r')
			{
				sbuf[spos++] = '\n';
				line++;

					// ignore following new line
				c = rbuf[rpos++];
				if (c == '\n')
					continue;
				else if (c == END_OF_DATA)
				{
					if ((len = reader.read(rbuf, 0, CHAR_READ_LENGTH)) < 0)
						throw new EOFException("EOF in whitespaces");
					rbuf[len] = END_OF_DATA;
					rpos = 0;
					c = rbuf[rpos++];
					if (c == '\n')
						continue;
				}
				rpos--;
			}
			else if (c == '\n')
			{
				sbuf[spos++] = c;
				line++;
			}
			else if (c == END_OF_DATA)
			{
				if ((len = reader.read(rbuf, 0, CHAR_READ_LENGTH)) < 0)
					throw new EOFException("EOF in whitespaces");
				rbuf[len] = END_OF_DATA;
				rpos = 0;
			}
			else
			{
					// end of whitespace
					// normal loop exit
				rpos--;
				break;
			}
		}

		// done with whitespaces
		//-------------------------------------------------------

			// is it an element or a value?
		if (rbuf[rpos] == '<')
		{
			rpos++;

			//=======================================================
			// Content is a collection of elements

			empty = true; // possibly empty
			for (;;)
			{
				c = rbuf[rpos];

					// break out on end tag
					// normal loop exit
				if (c == '/')
				{
						// report accumulated whitespaces inside empty element
					if (empty && spos != savePos)
					{
						docHandler.characters(saveBuf, savePos, spos - savePos);
						savePos = spos;
					}

					rpos++;
					break;
				}
				else if (c == END_OF_DATA)
				{
					if ((len = reader.read(rbuf, 0, CHAR_READ_LENGTH)) < 0)
						throw new EOFException("EOF in name");
					rbuf[len] = END_OF_DATA;
					rpos = 0;
					c = rbuf[rpos];
					if (c == '/')
					{
						if (empty && spos != savePos)
						{
							docHandler.characters(saveBuf, savePos, spos - savePos);
							savePos = spos;
						}
						rpos++;
						break;
					}
				}

					// definitely not empty
				empty = false;

				//=======================================================
				// Element, parse it by recursing into it

				readPos = rpos;
				parseElement();
				rpos = readPos;

				//=======================================================
				// Ignore whitespaces between elements

				for (;;)
				{
					c = rbuf[rpos];
					if (c == ' ' || c == '\t')
					{
						rpos++;
					}
					else if (c == '\r')
					{
						line++;
						cx = rbuf[++rpos];
						if (cx == '\n')
							rpos++;
						else if (cx == END_OF_DATA)
						{
							if ((len = reader.read(rbuf, 0, CHAR_READ_LENGTH)) < 0)
								throw new EOFException("EOF in whitespaces");
							rbuf[len] = END_OF_DATA;
							cx = rbuf[rpos = 0];
							if (cx == '\n')
								rpos++;
						}
					}
					else if (c == '\n')
					{
						rpos++;
						line++;
					}
					else if (c == END_OF_DATA)
					{
						if ((len = reader.read(rbuf, 0, CHAR_READ_LENGTH)) < 0)
							throw new EOFException("EOF within an element");

						rbuf[len] = END_OF_DATA;
						rpos = 0;
					}
					else
					{
							// end of whitespace
							// normal loop exit
						break;
					}
				}
				// passed through the whitespaces
				//-------------------------------------------------------

					// we should see next element now since
					// mixed content is not allowed.
				if (rbuf[rpos++] != '<')
					error("illegal character; expecting '<'");
			}
		}
		else	// CharData or CharRef
		{
			//=======================================================
			// Leaf node

				// report accumulated whitespaces, if any
			if (spos != savePos)
			{
				docHandler.characters(saveBuf, savePos, spos - savePos);
				savePos = spos;
			}

			// Scan CharData and CharRef until '<' is encountered

			spos = rpos;
			for (;;)
			{
				c = rbuf[rpos];
				if (   (c >  0x003E && c <= 0xD7FF)	// a-z, A-Z, punct, etc.
					|| (c <  0x0026 && c >= 0x0020)	// space & punct
					|| (c >  0x0026 && c <  0x003C)	// 0-9 & punct
					|| (c == 0x003D)
					|| (c == 0x0009)
					|| (c >= 0xE000 && c <= 0xFFFD)
					)
				{
					rpos++;
				}
				else if (c == '<')
				{
					// exit content on '<'
					// LOOP EXIT

					if (rpos != spos)
						docHandler.characters(readBuf, spos, rpos - spos);
					rpos++;
					break;
				}
				else if (c == '&')
				{
					if (rpos != spos)
						docHandler.characters(readBuf, spos, rpos - spos);
					rpos++;
					if (rbuf[rpos++] != '#')
					{
						if (rbuf[--rpos] == END_OF_DATA)
						{
							if ((len = reader.read(rbuf, 0, CHAR_READ_LENGTH)) < 0)
								throw new EOFException("EOF in numeric character reference");
							rbuf[len] = END_OF_DATA;
							rpos = 0;
							if (rbuf[rpos++] != '#')
								error("illegal character; expecting '#' in character reference");
						}
					}

					int code = 0;
					for (;;)
					{
							// the next character
						c = rbuf[rpos++];
						if (c >= '0' && c <= '9')
						{
							code *= 10;
							code += (c - '0');
						}
						else if (c == ';')
						{
							spos = rpos;
							break;
						}
						else if (c == END_OF_DATA)
						{
							if ((len = reader.read(rbuf, 0, CHAR_READ_LENGTH)) < 0)
								throw new EOFException("EOF in numeric character reference");
							rbuf[len] = END_OF_DATA;
							rpos = 0;
						}
						else
							error("illegal numeric character reference");
					}

					if (code <= 0xffff)
					{
						if ((code >= 0x0020 && code <= 0xD7FF) ||
							 code == 0x000A || code == 0x0009 || code == 0x000D ||
							(code >= 0xE000 && code <= 0xFFFD) ||
							(code >= 0x10000 && code <= 0x10ffff))
						{
							sbuf[TEMP_CHAR1] = (char)code;
							docHandler.characters(sbuf, TEMP_CHAR1, 1);
							continue;
						}
					}
					else if (code <= 0x0010ffff)
					{
						// we represent these as UNICODE surrogate pairs
						code -= 0x10000;
						sbuf[TEMP_CHAR1] = (char)(0xd800 | ((code >> 10) & 0x03ff));
						sbuf[TEMP_CHAR2] = (char)(0xdc00 | (code & 0x03ff));
						docHandler.characters(sbuf, TEMP_CHAR1, 2);
						continue;
					}
					else
					{
						readPos = rpos;
						error("unsupported unicode range");
					}
				}
				else if (c == '\r')
				{
					if (rpos != spos)
						docHandler.characters(readBuf, spos, rpos - spos);
					docHandler.characters(sbuf, TEMP_NEWLINE, 1);
					line++;

					cx = rbuf[++rpos];
					if (cx == '\n')
						rpos++;
					else if (cx == END_OF_DATA)
					{
						if ((len = reader.read(rbuf, 0, CHAR_READ_LENGTH)) < 0)
							throw new EOFException("EOF in whitespaces");
						rbuf[len] = END_OF_DATA;
						cx = rbuf[rpos = 0];
						if (cx == '\n')
							rpos++;
					}
					spos = rpos;
				}
				else if (c == '\n')
				{
					rpos++;
					line++;
				}
					// correctly paired surrogates are OK
				else if (c >= 0xd800 && c <= 0xdfff)
				{
						// flush
					if (rpos != spos)
						docHandler.characters(readBuf, spos, rpos - spos);
					cx = rbuf[++rpos];
					if (cx == END_OF_DATA)
					{
						if ((len = reader.read(rbuf, 0, CHAR_READ_LENGTH)) < 0)
							throw new EOFException("EOF in paired surrogate");
						rbuf[len] = END_OF_DATA;
						rpos = 0;
						spos = 0;
						cx = rbuf[rpos];
					}
					if (c >= 0xdc00 || cx < 0xdc00 || cx > 0xdfff)
					{
						readPos = rpos;
						error("illegal surrogate pair");
					}

					sbuf[TEMP_CHAR1] = c;
					sbuf[TEMP_CHAR2] = cx;
					docHandler.characters(sbuf, TEMP_CHAR1, 2);
					
					rpos++;
					spos = rpos;
				}
				else if (c == END_OF_DATA)
				{
					if (rpos != spos)
						docHandler.characters(readBuf, spos, rpos - spos);
					if ((len = reader.read(rbuf, 0, CHAR_READ_LENGTH)) < 0)
						throw new EOFException("EOF in character data");
					rbuf[len] = END_OF_DATA;
					rpos = 0;
					spos = 0;
				}
					// '>' must be escaped in textual content
				else if (c == '>')
				{
					readPos = rpos;
					error("unescaped '>' in textual content");
				}
				else
					error("illegal character in textual content");
			}

			if (rbuf[rpos++] != '/')
			{
				if (rbuf[--rpos] == END_OF_DATA)
				{
					if ((len = reader.read(rbuf, 0, CHAR_READ_LENGTH)) < 0)
						throw new EOFException("EOF before element");
					rbuf[len] = END_OF_DATA;
					rpos = 0;
				}
				if (rbuf[rpos++] != '/')
				{
					readPos = rpos - 1;
					error("illegal character; expecting end tag's '/'");
				}
			}
		}

			// make sure name matches in STag and ETag
	        // "</" was swallowed
		spos = name.start;
		int nend = name.start + name.count;
		while (spos < nend)
		{
			if (rbuf[rpos++] != sbuf[spos++])
			{
				if (rbuf[--rpos] == END_OF_DATA)
				{
					if ((len = reader.read(rbuf, 0, CHAR_READ_LENGTH)) < 0)
						throw new EOFException("EOF in end tag");
					rbuf[len] = END_OF_DATA;
					rpos = 0;
					spos--;
					continue;
				}
				else
				{
					readPos = rpos;
					error("STag and ETag mismatch");
				}
			}
		}

		c = rbuf[rpos++];
		if (c != '>')
		{
			if (c == END_OF_DATA)
			{
				if ((len = reader.read(rbuf, 0, CHAR_READ_LENGTH)) < 0)
					throw new EOFException("end of doc; expecting '>'");
				rbuf[len] = END_OF_DATA;
				rpos = 0;
				c = rbuf[rpos++];
			}
			if (c != '>')
			{
				readPos = rpos;
				error("illegal character; expecting end tag's '>'");
			}
		}
        docHandler.endElement(name.value);

		readPos = rpos;
    }

		private final void
	ignoreWhitespace(String fail)
		throws SAXException, IOException
	{
		// [9] WS ::= (#32 | #9 | #13 | #10)

		char[]	rbuf = readBuf;
		int		rpos = readPos;
		char	c;
		char	cx;
		int		len;

		for (;;)
		{
			c = rbuf[rpos];
			if (c == ' ' || c == '\t')
			{
				rpos++;
			}
			else if (c == '\r')
			{
				line++;
				cx = rbuf[++rpos];
				if (cx == '\n')
					rpos++;
				else if (cx == END_OF_DATA)
				{
					if ((len = reader.read(rbuf, 0, CHAR_READ_LENGTH)) < 0)
						throw new EOFException("EOF in whitespaces");
					rbuf[len] = END_OF_DATA;
					cx = rbuf[rpos = 0];
					if (cx == '\n')
						rpos++;
				}
			}
			else if (c == '\n')
			{
				rpos++;
				line++;
			}
			else if (c == END_OF_DATA)
			{
				if ((len = reader.read(rbuf, 0, CHAR_READ_LENGTH)) < 0)
					throw new EOFException("EOF within an element");

				rbuf[len] = END_OF_DATA;
				rpos = 0;
			}
			else
			{
					// end of whitespace
					// normal loop exit
				break;
			}
		}

		readPos = rpos;
	}

		private void
	error(String message)
		throws SAXException
    {
		closeReader();

		SAXParseException spe = new SAXParseException(message, this);
		errHandler.fatalError(spe);
		throw spe;
    }
}
