/*
 * Copyright 2000 by Docuverse
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Docuverse ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you
 * entered into with Docuverse.
 */

package com.docuverse.min;

//import java.io.CharConversionException;
import java.io.InputStream;
import java.io.IOException;
import java.io.Reader;

/**
Custom UTF-16 little-endian input stream reader optimized for
speed and size.  Methods not used by com.docuverse.min
parser are not supported.
*/
final class UTF16L extends Reader
{
	private static final int	END_OF_DATA = 0;  // EOD = double zero

	private InputStream		in;
	private boolean			closable;
	private byte[]			byteBuf;
	private int				bytePos;
	private int				readLen;	// read length (chunky read)

	UTF16L(InputStream in, boolean closable, byte[] buffer, int offset, int length)
	{
		super(in);

		this.in = in;
		this.closable = closable;
		this.byteBuf = buffer;
		this.bytePos = offset;
		this.readLen = buffer.length - (buffer.length % 1024);

		offset += length;
		buffer[offset++] = END_OF_DATA;
		buffer[offset] = END_OF_DATA;
	}

		public final void
	close() throws IOException
	{
		if (in != null)
		{
			if (closable)
				in.close();
			
			in = null;
		}
	}

		public final int
	read(char[] cb, int offset, int length)
		throws IOException
	{
		if (in == null)
			return -1;

		int		bi = bytePos;
		byte[]	bb = byteBuf;
		int		ce = offset + length;
		int		b1, b2;
		int		n;

		while (offset < ce)
		{
			b1 = bb[bi++];
			b2 = bb[bi++];
			if (b1 != 0 || b2 != 0)
			{
				cb[offset++] = (char)((b2 << 8) | b1);
				continue;
			}
			else
			{
				bi = 0;
				n = in.read(bb, 0, readLen);
				if (n > 0)
				{
					bb[n++] = END_OF_DATA;
					bb[n] = END_OF_DATA;
					continue;
				}
				else
				{
					bb[0] = END_OF_DATA;
					bb[1] = END_OF_DATA;
					close();
					break;
				}
			}
		}

		bytePos = bi;
		n = length - (ce - offset);
		if (n > 0)
			return n;

		return -1;
	}
}
