import java.io.*;
import java.util.Vector;
import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;

public abstract class MIDletBase extends MIDlet 
    implements CommandListener, Runnable {

    // commands
    Command exitCommand  = new Command("Exit", Command.EXIT, 2);
    Command testCommand = new Command("Test", Command.SCREEN, 1);
    Command checkCommand = new Command("Check", Command.SCREEN, 1);
    Command backToMainCommand = new Command("Back", Command.BACK, 1);
    Command backToResultCommand = new Command("Back", Command.BACK, 1);

    // UIs
    Display display;
    List fileList;
    TextBox resultBox;
    TextBox infoBox;
    TextBox checkBox;

    // data
    /** fileSpects[][] = { { filename, encoding, counts, comment }, ... } */
    protected String fileSpecs[][];
    int index = 0;

    boolean virgin = true;

    long warmupTime;
    long averageTime;
    long averageTimeNoGC;
    long averageTimeOnMem;
    long averageTimeNoGCOnMem;

    BenchBase benchmark;

    protected abstract BenchBase createBenchmark();

    //// extends MIDlet
    public void startApp() {
	if (display == null) {
	    display = Display.getDisplay(this);
	    benchmark = createBenchmark();
	    try {
		loadFileSpecs();
	    } catch (Exception e) {
		throw new RuntimeException(e.toString());
	    }
	}

	resultBox = new TextBox("test result:", "", 1024, 0);
	resultBox.setCommandListener(this);

	infoBox = new TextBox("file info:", "", 1024, 0);
	infoBox.addCommand(backToMainCommand);
	infoBox.setCommandListener(this);

	checkBox = new TextBox("check result:", "", 1024, 0);
	checkBox.addCommand(backToResultCommand);
	checkBox.setCommandListener(this);

	fileList = new List("select a file:", List.IMPLICIT);
	for (int i=0; i<fileSpecs.length; i++) {
	    fileList.append(fileSpecs[i][0], null);
	}
	fileList.addCommand(exitCommand);
	fileList.addCommand(testCommand);
	fileList.setCommandListener(this);

	display.setCurrent(fileList);
    }
    
    public void pauseApp() {}
    public void destroyApp(boolean unconditional) {}

    //// implements CommandListener
    public void commandAction(Command c, Displayable s) {
	if (c == exitCommand) {
	    destroyApp(false);
	    notifyDestroyed();

	} else if (c == testCommand) {
	    index = fileList.getSelectedIndex();
	    resultBox.setString("wait...");
	    resultBox.removeCommand(backToMainCommand);
	    resultBox.removeCommand(checkCommand);
	    display.setCurrent(resultBox);
	    display.callSerially(this); //run();

	} else if (c == backToMainCommand) {
	    display.setCurrent(fileList);
	    fileList.setSelectedIndex(index, true);

	} else if (c == backToResultCommand) {
	    display.setCurrent(resultBox);

	} else if (c == checkCommand) {
	    String correct = fileSpecs[index][2];
	    String result = countToString();
	    String out = "correct:\n  " + correct + "\n"
		+ "result:\n  " + result + "\n\n"
		+ (result.equals(correct) ? "OK." : "NG");
	    checkBox.setString(out);
	    display.setCurrent(checkBox);

	} else if (c == List.SELECT_COMMAND) {
	    index = fileList.getSelectedIndex();
	    infoBox.setTitle(fileSpecs[index][0]);
	    String out = fileSpecs[index][3] + "\n"
		+ "encoding:" + fileSpecs[index][1] + "\n"
		+ "cnt:" + fileSpecs[index][2];
	    infoBox.setString(out);
	    display.setCurrent(infoBox);
	}
    }

    //// implements Runnable
    public void run() {
	String filename = fileSpecs[index][0];
	String encoding = fileSpecs[index][1];
	String out = "file:" + filename + "\n";
	try {
	    runBenchmark(filename, encoding);
	    out += "cnt:" + countToString() + "\n";
	    if (virgin) {
		out += "1st:" + warmupTime + "ms" + "\n";
		virgin = false;
	    }
	    //out += "avg:" + averageTime + "ms" + "\n";
	    out += "ngc:" + averageTimeNoGC + "ms" + "\n";
	    //out += "avgM:" + averageTimeOnMem + "ms" + "\n";
	    //out += "ngcM:" + averageTimeNoGCOnMem + "ms";
	} catch (Exception ex) {
	    out = "ERROR:" + ex.toString();
	    ex.printStackTrace();
	}
	resultBox.setString(out);
	resultBox.addCommand(backToMainCommand);
	resultBox.addCommand(checkCommand);
	display.setCurrent(resultBox);
    }

    ////////////////////////////////////////////////////////////////////////
    protected void runBenchmark(String filename, String encoding)
	throws Exception {
	benchmark.init(filename, encoding, false);
	warmupTime = benchmark.runWarmup();
	//averageTime = benchmark.runNormal(5);
	averageTimeNoGC = benchmark.runNoGC(5);

	//benchmark.init(filename, encoding, true);
	//averageTimeOnMem = benchmark.runNormal(5);
	//averageTimeNoGCOnMem = benchmark.runNoGC(5);
    }

    protected String countToString() {
	return benchmark.elementCount 
	    + "/" + benchmark.attributeCount 
	    + "/" + benchmark.characterCount;
    }

    protected void loadFileSpecs() throws Exception {
	InputStream s = getClass().getResourceAsStream("dir.txt");
	// line format of dir.txt:
	//   filename-1 encoding elms/atts/chars comment-in-english
	// filename must not have space character.
	//
	InputStreamReader reader = new InputStreamReader(s, "US-ASCII");
	boolean eol = false;
	Vector v = new Vector();
	StringBuffer buf = new StringBuffer();
	int c;
	int f = 0;
	String spec[] = new String[4];
	do {
	    c =  reader.read();
	    if (c == '\r' || c == '\n'|| c == -1) {
		if (eol) {
		    continue;
		} else {
		    eol = true;
		    spec[f] = buf.toString();
		    v.addElement(spec);
		    f = 0;
		    buf.setLength(0);
		    spec = new String[4];
		} 
	    } else {
		eol = false;
		if ((c == ' ' || c == '\t') && f < 3) {
		    spec[f] = buf.toString();
		    buf.setLength(0);
		    f++;
		} else {
		    buf.append((char)c);
		}
	    }
	} while(c != -1);
	fileSpecs = new String[v.size()][];
	for (int i=0; i<v.size(); i++) {
	    fileSpecs[i] = (String[])v.elementAt(i);
	    //System.err.println(fileSpecs[i][0]);
	}
    }
}
